// NOTE: DO NOT CHANGE THIS FILE. IT IS AUTOMATICALLY GENERATED.

#include <mbgl/shaders/clipping_mask.hpp>
#include <mbgl/shaders/source.hpp>

namespace mbgl {
namespace shaders {

const char* clipping_mask::name = "clipping_mask";
const char* clipping_mask::vertexSource = source() + 10630;
const char* clipping_mask::fragmentSource = source() + 10742;

} // namespace shaders
} // namespace mbgl
