// NOTE: DO NOT CHANGE THIS FILE. IT IS AUTOMATICALLY GENERATED.

#include <mbgl/shaders/extrusion_texture.hpp>
#include <mbgl/shaders/source.hpp>

namespace mbgl {
namespace shaders {

const char* extrusion_texture::name = "extrusion_texture";
const char* extrusion_texture::vertexSource = source() + 31747;
const char* extrusion_texture::fragmentSource = source() + 31963;

} // namespace shaders
} // namespace mbgl
