// NOTE: DO NOT CHANGE THIS FILE. IT IS AUTOMATICALLY GENERATED.

#include <mbgl/shaders/heatmap.hpp>
#include <mbgl/shaders/source.hpp>

namespace mbgl {
namespace shaders {

const char* heatmap::name = "heatmap";
const char* heatmap::vertexSource = source() + 10789;
const char* heatmap::fragmentSource = source() + 13245;

} // namespace shaders
} // namespace mbgl
