// NOTE: DO NOT CHANGE THIS FILE. IT IS AUTOMATICALLY GENERATED.

#include <mbgl/shaders/background_pattern.hpp>
#include <mbgl/shaders/source.hpp>

namespace mbgl {
namespace shaders {

const char* background_pattern::name = "background_pattern";
const char* background_pattern::vertexSource = source() + 3019;
const char* background_pattern::fragmentSource = source() + 3663;

} // namespace shaders
} // namespace mbgl
