// NOTE: DO NOT CHANGE THIS FILE. IT IS AUTOMATICALLY GENERATED.

#include <mbgl/shaders/line.hpp>
#include <mbgl/shaders/source.hpp>

namespace mbgl {
namespace shaders {

const char* line::name = "line";
const char* line::vertexSource = source() + 37744;
const char* line::fragmentSource = source() + 42206;

} // namespace shaders
} // namespace mbgl
