// NOTE: DO NOT CHANGE THIS FILE. IT IS AUTOMATICALLY GENERATED.

#include <mbgl/shaders/hillshade.hpp>
#include <mbgl/shaders/source.hpp>

namespace mbgl {
namespace shaders {

const char* hillshade::name = "hillshade";
const char* hillshade::vertexSource = source() + 34909;
const char* hillshade::fragmentSource = source() + 35108;

} // namespace shaders
} // namespace mbgl
