// NOTE: DO NOT CHANGE THIS FILE. IT IS AUTOMATICALLY GENERATED.

#include <mbgl/shaders/fill_extrusion.hpp>
#include <mbgl/shaders/source.hpp>

namespace mbgl {
namespace shaders {

const char* fill_extrusion::name = "fill_extrusion";
const char* fill_extrusion::vertexSource = source() + 24858;
const char* fill_extrusion::fragmentSource = source() + 27694;

} // namespace shaders
} // namespace mbgl
