/****************************************************************************
**
** Copyright (C) 2016 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of the test suite of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:GPL-EXCEPT$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3 as published by the Free Software
** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
** included in the packaging of this file. Please review the following
** information to ensure the GNU General Public License requirements will
** be met: https://www.gnu.org/licenses/gpl-3.0.html.
**
** $QT_END_LICENSE$
**
****************************************************************************/

#ifndef QGEOROUTINGMANAGERENGINE_TEST_H
#define QGEOROUTINGMANAGERENGINE_TEST_H

#include <qgeoserviceprovider.h>
#include <qgeoroutingmanagerengine.h>
#include <qgeocoordinate.h>
#include <qgeoroutereply.h>
#include <qgeorouterequest.h>

QT_USE_NAMESPACE

class QGeoRoutingManagerEngineTest: public QGeoRoutingManagerEngine

{
Q_OBJECT
public:
    QGeoRoutingManagerEngineTest(const QVariantMap &parameters,
        QGeoServiceProvider::Error *error, QString *errorString) :
        QGeoRoutingManagerEngine(parameters)
    {
        Q_UNUSED(error);
        Q_UNUSED(errorString);
        setSupportedTravelModes(QGeoRouteRequest::PedestrianTravel);
        setSupportedFeatureTypes(QGeoRouteRequest::TollFeature);
        setSupportedFeatureWeights(QGeoRouteRequest::PreferFeatureWeight);
        setSupportedRouteOptimizations(QGeoRouteRequest::FastestRoute);
        setSupportedSegmentDetails(QGeoRouteRequest::BasicSegmentData);
        setSupportedManeuverDetails(QGeoRouteRequest::BasicManeuvers);
    }

    QGeoRouteReply* calculateRoute(const QGeoRouteRequest& request)
    {
        Q_UNUSED(request);
        return new QGeoRouteReply(QGeoRouteReply::NoError,"no error");
    }

    QGeoRouteReply* updateRoute(const QGeoRoute &route, const QGeoCoordinate &position)
    {
        Q_UNUSED(route);
        Q_UNUSED(position);
        return new QGeoRouteReply(QGeoRouteReply::CommunicationError,"no error");

    }


};

#endif
