// NOTE: DO NOT CHANGE THIS FILE. IT IS AUTOMATICALLY GENERATED.

#include <mbgl/shaders/raster.hpp>
#include <mbgl/shaders/source.hpp>

namespace mbgl {
namespace shaders {

const char* raster::name = "raster";
const char* raster::vertexSource = source() + 57390;
const char* raster::fragmentSource = source() + 58203;

} // namespace shaders
} // namespace mbgl
