// This file is generated. Edit scripts/generate-style-code.js, then run `make style-code`.

#include <mbgl/style/layers/circle_layer_properties.hpp>

namespace mbgl {
namespace style {

} // namespace style
} // namespace mbgl
