/* This file was generated by syncqt. */
#ifndef QT_QTPOSITIONING_VERSION_H
#define QT_QTPOSITIONING_VERSION_H

#define QTPOSITIONING_VERSION_STR "5.15.17"

#define QTPOSITIONING_VERSION 0x050F11

#endif // QT_QTPOSITIONING_VERSION_H
