#pragma once

/// The major version number
#define VECTOR_TILE_VERSION_MAJOR 1

/// The minor version number
#define VECTOR_TILE_VERSION_MINOR 0

/// The patch number
#define VECTOR_TILE_VERSION_PATCH 2

/// The complete version number
#define VECTOR_TILE_VERSION_CODE (VECTOR_TILE_VERSION_MAJOR * 10000 + VECTOR_TILE_VERSION_MINOR * 100 + VECTOR_TILE_VERSION_PATCH)

/// Version number as string
#define VECTOR_TILE_VERSION_STRING "1.0.2"
