// NOTE: DO NOT CHANGE THIS FILE. IT IS AUTOMATICALLY GENERATED.

#include <mbgl/shaders/fill_pattern.hpp>
#include <mbgl/shaders/source.hpp>

namespace mbgl {
namespace shaders {

const char* fill_pattern::name = "fill_pattern";
const char* fill_pattern::vertexSource = source() + 22998;
const char* fill_pattern::fragmentSource = source() + 23950;

} // namespace shaders
} // namespace mbgl
