// NOTE: DO NOT CHANGE THIS FILE. IT IS AUTOMATICALLY GENERATED.

#include <mbgl/shaders/collision_circle.hpp>
#include <mbgl/shaders/source.hpp>

namespace mbgl {
namespace shaders {

const char* collision_circle::name = "collision_circle";
const char* collision_circle::vertexSource = source() + 15668;
const char* collision_circle::fragmentSource = source() + 16974;

} // namespace shaders
} // namespace mbgl
