#pragma once

namespace mbgl {
namespace style {

enum class LayerType {
    Fill,
    Line,
    Circle,
    Symbol,
    Raster,
    Hillshade,
    Background,
    Custom,
    FillExtrusion,
    Heatmap,
};

} // namespace style
} // namespace mbgl
