// NOTE: DO NOT CHANGE THIS FILE. IT IS AUTOMATICALLY GENERATED.

#include <mbgl/shaders/preludes.hpp>
#include <mbgl/shaders/source.hpp>

namespace mbgl {
namespace shaders {

const char* vertexPrelude = source() + 0;
const char* fragmentPrelude = source() + 2557;

} // namespace shaders
} // namespace mbgl
