// NOTE: DO NOT CHANGE THIS FILE. IT IS AUTOMATICALLY GENERATED.

#include <mbgl/shaders/background.hpp>
#include <mbgl/shaders/source.hpp>

namespace mbgl {
namespace shaders {

const char* background::name = "background";
const char* background::vertexSource = source() + 2738;
const char* background::fragmentSource = source() + 2850;

} // namespace shaders
} // namespace mbgl
